//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 05: OpenMp
//
//  Description: Sort in link lists using openMP and qsort
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v: gcc version 9.2.1 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS:     Linux Ubuntu 19.10
CPU:    8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************

Compiling:	gcc omplinkSort02.c -o omplinkSort02 -fopenmp -O3
Executing:	./omplinkSort02

Output in my Laptop:

real	3m5.320s
user	3m7.042s
sys	0m1.224s

*/
//************************************************************


// *******************************************************************
//#pragma GCC optimize("O3","unroll-loops","omit-frame-pointer","inline")
//#pragma GCC option("arch=native","tune=native","no-zero-upper")
//************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>

// ******************************************
// Array Size
#define Nnodes     1000
#define N          100000

typedef struct node {
	unsigned short int A[N] ;
	struct node *next ;
	} NODE ;

NODE *Top ;


// **************************************************************
unsigned int seed = 666999666 ;  // random number seed

// **************************************************************
unsigned int randUint( void )
{
#pragma omp threadprivate(seed)
	seed = seed * 1103515245 + 12345;
return seed ;
}


// ***************************************
int compar(const void *a, const void *b)
{
if ( *(unsigned short *)a >  *(unsigned short *)b )
	return 1 ;
if ( *(unsigned short *)a <  *(unsigned short *)b )
	return -1 ;
else
	return 0 ;
}


// ***************************************
NODE *createLinkList( void )
{
NODE *ptop, *pcur ;

ptop = malloc(sizeof(NODE)) ;
if ( ptop == NULL )
	return NULL ;
pcur = ptop ;
for (unsigned int i=1 ; i < Nnodes ; i++ ) {
	pcur->next = malloc(sizeof(NODE)) ;
	if ( pcur->next == NULL )
		return NULL ;
	pcur=pcur->next ;
	}
pcur->next = NULL ;

pcur = ptop ;

do {
#pragma omp parallel for
	for( int j = 0 ; j < N ; j++ )
		pcur->A[j] = randUint() % 65536 ;
	pcur = pcur->next ;
} while( pcur != NULL ) ;

return(ptop) ;
}

// ***************************************
void doSortInList( NODE *top )
{
while (top) {
	qsort( top->A, N, sizeof(unsigned short), compar ) ;
	top = top->next ;
	}
}


// ***************************************
int main()
{

#pragma omp threadprivate(seed)

Top = createLinkList() ;
doSortInList( Top ) ;

return 0 ;
}
