//************************************************************
//  University of Patras
//	Department of Electrical Engineering & Computer Technology
//
//	Course: Parallel Programming in AI
//	Week 05: OpenMp
//
//  Description: Sort in link lists using openMP and qsort
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v: g++ version 9.2.1 
OS:     Linux Ubuntu 19.10
CPU:    8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************

Compiling:	g++ omplinkSort03.c -o omplinkSort03 -fopenmp
Executing:	./omplinkSort03

Output in my Laptop:

real	0m35.883s
user	4m43.225s
sys	0m1.372s

*/
//************************************************************



#include <iostream>
#include <algorithm>
#include <omp.h>

// ******************************************
// Array Size
#define Nnodes     1000
#define N          100000

typedef struct node {
	unsigned short int A[N] ;
	struct node *next ;
	} NODE ;

NODE *Top ;


// **************************************************************
unsigned int seed = 666999666 ;  // random number seed

// **************************************************************
unsigned int randUint( void )
{
#pragma omp threadprivate(seed)
	seed = seed * 1103515245 + 12345;
return seed ;
}


// ***************************************
int compar(const void *a, const void *b)
{
if ( *(unsigned short *)a >  *(unsigned short *)b )
	return 1 ;
if ( *(unsigned short *)a <  *(unsigned short *)b )
	return -1 ;
else
	return 0 ;
}


// ***************************************
NODE *createLinkList( void )
{
NODE *ptop, *pcur ;

ptop = new NODE ;
if ( ptop == NULL )
	return NULL ;
pcur = ptop ;
for (unsigned int i=1 ; i < Nnodes ; i++ ) {
	pcur->next = new NODE ;
	if ( pcur->next == NULL )
		return NULL ;
	pcur=pcur->next ;
	}
pcur->next = NULL ;

pcur = ptop ;
do {
#pragma omp parallel for
	for( int j = 0 ; j < N ; j++ )
		pcur->A[j] = randUint() % 65536 ;
	pcur = pcur->next ;
} while( pcur != NULL ) ;

return(ptop) ;
}

// ***************************************
void doSortInList( NODE *top )
{
#pragma omp parallel
{
#pragma omp master
{
while (top) {
#pragma omp task firstprivate(top)
	{
		std::sort( top->A, top->A+N ) ;
	}
	top = top->next ;
	}
}
}
}


// ***************************************
int main()
{

#pragma omp threadprivate(seed)

Top = createLinkList() ;
doSortInList( Top ) ;

return 0 ;
}
